/* data_types.h
 * 
 * Copyright (C) 2015 Tristan Dagobert, Ecole Normale Superieure de Cachan.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or 
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @mainpage
 *
 * README.txt:
 * @verbinclude README.txt
 */

/**
 * @file data_types.h
 * @brief
 *
 * 
 *
 * @todo 
 *
 * @author Tristan Dagobert <tristan.dagobert@cmla.ens-cachan.fr>
 */

#if !(defined _COMPOSANTS_H_)
#define _COMPOSANTS_H_

#define FIRST_COMPONENT_ALGO    0
#define PARTIAL_RECOMPOS_ALGO   1
#define SELECTED_COMPONENT_ALGO 2

#define WITH_REQUANTIZATION 0
#define PERMIT_SATURATION 1
#define NO_REQUANTIZATION 2

/* linked list containing the images of the
 * sliding window */
struct window_s
{
  float *im;
  struct window_s *next;
};
typedef struct window_s window;

/* API containing the sliding window and its
 * properties */
struct window_API_s
{
  unsigned int nrows; /** number of rows of each image */
  unsigned int ncols; /** number of columns of each image */
  unsigned int npixs; /** number of pixels of each image */
  unsigned int nimages;	/** number of images into the window */

  int type;
  window *win; /** linked list of the images */
  window *win_end; /** pointer to the last element of list */
};
typedef struct window_API_s window_API;

/* double linked list containing the file
 * names of image */
struct files_list_s
{
  char *file_name;
  struct files_list_s *next;
  struct files_list_s *prev;
};
typedef struct files_list_s files_list;

/* structure relative to the command line
 * parameters */
struct parameters_s
{
  char *list_file_name;
  char *output_file_name;
  unsigned int window_width;
  unsigned int first_eigenvec;
  unsigned int last_eigenvec;
  float epsilon;
  int algorithm;
};
typedef struct parameters_s parameters;

#endif
