/* pca_method.c
 * 
 * Copyright (C) 2015 Tristan Dagobert, Ecole Normale Superieure de Cachan.
 * Copyright (C) 1996, 1997, 1998, 1999, 2000, 2006, 2007 Gerard Jungman, 
 * Patrick Alken and B. Gough for function sort_eigens().
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or 
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @mainpage
 *
 * README.txt:
 * @verbinclude README.txt
 */

/**
 * @file pca_method.c
 * @brief
 *
 * 
 *
 * @todo 
 *
 * @author Tristan Dagobert <tristan.dagobert@cmla.ens-cachan.fr>
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <cblas.h>

#include "data_types.h"
#include "io_auxiliary.h"
#include "pca_method.h"
#include "ccmath_svd.h"

float * compute_laplacian_mu(float * mu, float * laplacian, window_API * win_API)
{
 float * laplacian_mu = NULL;
 int ncol, nrow;
 float a11, a12, a13, a21, a22, a23, a31, a32, a33;

 ncol = win_API->ncols;
 nrow = win_API->nrows;

 laplacian_mu = (float *) calloc (ncol * nrow, sizeof (float));
 if (laplacian_mu == NULL)
    {
      printf ("allocation failed\n");
      exit (EXIT_FAILURE);
    }
 int i,j;

 for (i=0; i<nrow; i++)
   {
     for(j=0; j<ncol; j++)
       {
	 /* corner cases */
	 if (i==0)
	   {
	     if (j==0)
	       {
		 a11 = mu[i*ncol + j];
		 a12 = mu[i*ncol + j];
		 a13 = mu[i*ncol + j + 1];
		 
		 a21 = mu[i*ncol + j];
		 a22 = mu[i*ncol + j];
		 a23 = mu[i*ncol + j + 1];

		 a31 = mu[(i+1)*ncol + j];
		 a32 = mu[(i+1)*ncol + j];
		 a33 = mu[(i+1)*ncol + j + 1];
	       }
	     else if (j==ncol-1)
	       {
		 a11 = mu[i*ncol + j - 1];
		 a12 = mu[i*ncol + j];
		 a13 = mu[i*ncol + j];
		 
		 a21 = mu[i*ncol + j - 1];
		 a22 = mu[i*ncol + j];
		 a23 = mu[i*ncol + j];

		 a31 = mu[(i+1)*ncol + j - 1];
		 a32 = mu[(i+1)*ncol + j];
		 a33 = mu[(i+1)*ncol + j];
	       }
	     else
	       {
		 a11 = mu[i*ncol + j - 1];
		 a12 = mu[i*ncol + j];
		 a13 = mu[i*ncol + j + 1];
		 
		 a21 = mu[i*ncol + j - 1];
		 a22 = mu[i*ncol + j];
		 a23 = mu[i*ncol + j + 1];

		 a31 = mu[(i+1)*ncol + j - 1];
		 a32 = mu[(i+1)*ncol + j];
		 a33 = mu[(i+1)*ncol + j + 1];
	       }
	   }
	 else if (i==nrow-1)
	   {
	     if (j==0)
	       {
		 a11 = mu[(i-1)*ncol + j];
		 a12 = mu[(i-1)*ncol + j];
		 a13 = mu[(i-1)*ncol + j + 1];
		 
		 a21 = mu[i*ncol + j];
		 a22 = mu[i*ncol + j];
		 a23 = mu[i*ncol + j + 1];

		 a31 = mu[i*ncol + j];
		 a32 = mu[i*ncol + j];
		 a33 = mu[i*ncol + j + 1];
	       }
	     else if (j==ncol-1)
	       {
		 a11 = mu[(i-1)*ncol + j - 1];
		 a12 = mu[(i-1)*ncol + j];
		 a13 = mu[(i-1)*ncol + j];
		 
		 a21 = mu[i*ncol + j - 1];
		 a22 = mu[i*ncol + j];
		 a23 = mu[i*ncol + j];

		 a31 = mu[i*ncol + j - 1];
		 a32 = mu[i*ncol + j];
		 a33 = mu[i*ncol + j];		 
	       }
	     else
	       {		 
		 a11 = mu[(i-1)*ncol + j - 1];
		 a12 = mu[(i-1)*ncol + j];
		 a13 = mu[(i-1)*ncol + j + 1];
		 
		 a21 = mu[i*ncol + j - 1];
		 a22 = mu[i*ncol + j];
		 a23 = mu[i*ncol + j + 1];

		 a31 = mu[i*ncol + j - 1];
		 a32 = mu[i*ncol + j];
		 a33 = mu[i*ncol + j + 1];
	       }
	   }
	 else
	   {	
	     if (j==0)
	       {
		 a11 = mu[(i-1)*ncol + j];
		 a12 = mu[(i-1)*ncol + j];
		 a13 = mu[(i-1)*ncol + j + 1];
		 
		 a21 = mu[i*ncol + j];
		 a22 = mu[i*ncol + j];
		 a23 = mu[i*ncol + j + 1];

		 a31 = mu[(i+1)*ncol + j];
		 a32 = mu[(i+1)*ncol + j];
		 a33 = mu[(i+1)*ncol + j + 1];
	       }
	     else if (j==ncol-1)
	       {
		 a11 = mu[(i-1)*ncol + j - 1];
		 a12 = mu[(i-1)*ncol + j];
		 a13 = mu[(i-1)*ncol + j];
		 
		 a21 = mu[i*ncol + j - 1];
		 a22 = mu[i*ncol + j];
		 a23 = mu[i*ncol + j];

		 a31 = mu[(i+1)*ncol + j - 1];
		 a32 = mu[(i+1)*ncol + j];
		 a33 = mu[(i+1)*ncol + j];		 
	       }
	     else
	       {		 
		 a11 = mu[(i-1)*ncol + j - 1];
		 a12 = mu[(i-1)*ncol + j];
		 a13 = mu[(i-1)*ncol + j + 1];
		 
		 a21 = mu[i*ncol + j - 1];
		 a22 = mu[i*ncol + j];
		 a23 = mu[i*ncol + j + 1];

		 a31 = mu[(i+1)*ncol + j - 1];
		 a32 = mu[(i+1)*ncol + j];
		 a33 = mu[(i+1)*ncol + j + 1];
	       }

	   }
	 laplacian_mu[i*ncol + j] =
	   a11 * laplacian[0] + a12 * laplacian[1] + a13 * laplacian[2]
	   +
	   a21 * laplacian[3] + a22 * laplacian[4] + a23 * laplacian[5]
	   +
	   a31 * laplacian[6] + a32 * laplacian[7] + a33 * laplacian[8];

       }
   }
 return(laplacian_mu);
}

float * create_laplacian_filter() 
{
  float * laplacian = NULL;

  /* allocation */
  laplacian = (float *) calloc (9, sizeof (float));
  if (laplacian == NULL)
    {
      perror ("create_laplacian ()\n");
      exit (EXIT_FAILURE);
    }
  laplacian[0] = 1;
  laplacian[1] = 1;
  laplacian[2] = 1;

  laplacian[3] = 1;
  laplacian[4] = -8;
  laplacian[5] = 1;

  laplacian[6] = 1;
  laplacian[7] = 1;
  laplacian[8] = 1;

  return(laplacian);
}

/**
 * @brief compute the mean image of the window
 *
 * @param win_API the window API
 * @return the mean image
 */
float *
compute_mean_image (window_API * win_API)
{
  window *win = NULL;
  float *mean_im = NULL;
  int i = 0;
  int npixs = win_API->npixs;
  int N = win_API->nimages;

  /* allocation */
  mean_im = (float *) calloc (npixs, sizeof (float));
  if (mean_im == NULL)
    {
      printf ("allocation failed\n");
      exit (EXIT_FAILURE);
    }

  /* sum */
  win = win_API->win;
  while (win != NULL)
    {
      for (i = 0; i < npixs; i++)
	{
	  mean_im[i] += win->im[i];
	}
      win = win->next;
    }
  /* mean */
  for (i = 0; i < npixs; i++)
    {
      mean_im[i] /= N;
    }

  return mean_im;
}

/** 
 * @brief create the matrix A
 *        The images window is converted into a data vector A of floats,
 *        in which images are stocked columns by columns ones under others
 *
 * @param win_API the images window
 * @param mean_im the mean image 
 * @return matrix A
 */
float *
create_matrix_A (window_API * win_API, float *mean_im)
{
  window *win;
  float *A;
  int i;
  int npixs, N;
  float *Aj;

  npixs = win_API->npixs;
  N = win_API->nimages;

  /* appropriate allocation */
  A = (float *) malloc (npixs * N * sizeof (float));
  if (A == NULL)
    {
      printf ("allocation failed\n");
      exit (EXIT_FAILURE);
    }

  /* appropriate copy */
  win = win_API->win;
  for (i = 0; i < N; i++)
    {
      memcpy (A + (i * npixs), win->im, npixs * sizeof (float));
      win = win->next;
    }

  /* center values with the mean image */
  for (i = 0; i < N; i++)
    {
      Aj = A + (i * npixs);
      cblas_saxpy (npixs, -1.0, mean_im, 1, Aj, 1);
    }

  return A;
}

/**
 * @brief compute the eigen values and vectors of the matrix At.A
 *        where At is the A transpose matrix. 
 * @param A the matrix A
 * @param A_nrows rows number of A which corresponds to the size of images 
 * @param A_ncols colums number of A which corresponds to the window length
 * @return a data vector representing the eigen vectors matrix 
 * sorted according the decreasingly eigen values amplitude
 */
float *
compute_eigens (float *A, int A_nrows, int A_ncols)
{
  int i = 0, j = 0;
  float *At_A = NULL;
  float *eigen_vecs_t = NULL, *eigen_vecs = NULL, *eigen_vals = NULL;
  float *u_dummy;

  /* matrix At.A allocation */
  At_A = (float *) malloc (A_ncols * A_ncols * sizeof (float));
  if (At_A == NULL)
    {
      printf ("allocation failed\n");
      exit (EXIT_FAILURE);
    }

  /* matrix product At x A using CBLAS library */
  cblas_sgemm (CblasColMajor, CblasTrans, CblasNoTrans,
	       A_ncols, A_ncols, A_nrows, 1.0, A, A_nrows, A, A_nrows, 0.0,
	       At_A, A_ncols);

  /* transposed matrix of eigen vectors */
  eigen_vecs_t = (float *) calloc (A_ncols * A_ncols, sizeof (float));
  if (eigen_vecs_t == NULL)
    {
      printf ("allocation failed\n");
      exit (EXIT_FAILURE);
    }

  /* matrix of eigen vectors */
  eigen_vecs = (float *) calloc (A_ncols * A_ncols, sizeof (float));
  if (eigen_vecs == NULL)
    {
      printf ("allocation failed\n");
      exit (EXIT_FAILURE);
    }

  /* u matrix of the SVD decomposition, but useless in this case */
  u_dummy = (float *) calloc (A_ncols * A_ncols, sizeof (float));
  if (u_dummy == NULL)
    {
      printf ("allocation failed\n");
      exit (EXIT_FAILURE);
    }

  /* eigen values allocation */
  eigen_vals = (float *) calloc (A_ncols, sizeof (float));
  if (eigen_vals == NULL)
    {
      printf ("allocation failed\n");
      exit (EXIT_FAILURE);
    }

  /* SVD computation */
  svd_1 (eigen_vals, At_A, u_dummy, A_ncols, eigen_vecs_t, A_ncols);
  free (u_dummy);

  /* because previous function gives the transposed matrix of eigen vectors,
     we have to build the matrix eigen vectors */
  for (i = 0; i < A_ncols; i++)
    {
      /* for each vector v(i), we pick its components */
      for (j = 0; j < A_ncols; j++)
	{
	  eigen_vecs[i * A_ncols + j] = eigen_vecs_t[i + j * A_ncols];
	}
    }

  /* descending sort according eigen values amplitude */
  sort_eigens (eigen_vals, eigen_vecs, A_ncols);
#if defined(EXPLICIT_MODE)
  print_matrix ("eigenvalues_sorted", 1, A_ncols, eigen_vals, 1);
  print_eigenvalues (eigen_vals, A_ncols);
#endif

  /* libration conditionnelle */
  free (At_A);
  free (eigen_vals);
  free (eigen_vecs_t);

  return (eigen_vecs);
}

/**
 * @brief Sort the eigen values and vectors in the decreasing order
 *        depending on the eigen values amplitude.
 *
 * @param eigen_vals eigen values
 * @param eigen_vecs eigen vectors
 * @param neigens number of eigen values
 */
void
sort_eigens (float *eigen_vals, float *eigen_vecs, int neigens)
{
  int i, j, k;
  float ek, ej, val_tmp;
  float *vec_tmp;

  vec_tmp = (float *) malloc (sizeof (float) * neigens);
  if (vec_tmp == NULL)
    {
      printf ("allocation failed\n");
      exit (EXIT_FAILURE);
    }

  for (i = 0; i < neigens - 1; i++)
    {
      k = i;
      ek = eigen_vals[i];

      /* search for something to permute */
      for (j = i + 1; j < neigens; j++)
	{
	  ej = eigen_vals[j];

	  if (fabs (ej) > fabs (ek))
	    {
	      k = j;
	      ek = ej;
	    }
	}

      if (k != i)
	{
	  /* eigenvalues permutation */
	  val_tmp = eigen_vals[i];
	  eigen_vals[i] = eigen_vals[k];
	  eigen_vals[k] = val_tmp;

	  /* eigenvectors permutation */
	  memcpy (vec_tmp, eigen_vecs + (i * neigens),
		  neigens * sizeof (float));
	  memcpy (eigen_vecs + (i * neigens), eigen_vecs + (k * neigens),
		  neigens * sizeof (float));
	  memcpy (eigen_vecs + (k * neigens), vec_tmp,
		  neigens * sizeof (float));
	}
    }
  free (vec_tmp);
  return;
}

/**
 * @brief Compute the orthonormal basis of eigenvectors w(i) of matrix A.At 
 * given by the relation :
 *        w(i) = A.v(i)/ ||A.v(i)||
 * @param A matrix A
 * @param A_nrows rows number of A 
 * @param A_ncols columns number of A
 * @param eigen_vecs eigenvectors v(i) previously computed and sorted
 * @return a data vector representing the eigenvectors matrix w(i) 
 */
float *
compute_basis (float *A, int A_nrows, int A_ncols, float *eigen_vecs)
{
  int i = 0;
  /*int j = 0; */
  float *basis;
  float norme = 0.0;		/* a=0.0; */

  basis = (float *) calloc (A_nrows * A_ncols, sizeof (float));
  if (basis == NULL)
    {
      printf ("allocation failed\n");
      exit (EXIT_FAILURE);
    }

  /* for each eigen vectors v(i) of At.A, we construct
   * the eigen vectors w(i) of A.At which is related
   * by w(i) = A.v(i)
   */
  for (i = 0; i < A_ncols; i++)
    {
      /* product A x v(i) */
      cblas_sgemv (CblasColMajor, CblasNoTrans, A_nrows, A_ncols, 1.0, A,
		   A_nrows, eigen_vecs + (i * A_ncols), 1, 0.0,
		   basis + (i * A_nrows), 1);
    }

  /* we normalize w(i) */
  int k;
  for (i = 0; i < A_ncols - 1; i++)
    {
      norme = 0.0;
      norme =
	cblas_sdot (A_nrows, basis + i * A_nrows, 1, basis + i * A_nrows, 1);
      norme = sqrt (norme);

      // normalization     
      for (k = 0; k < A_nrows; k++)
	{
	  *(basis + i * A_nrows + k) /= norme;
	}
    }

  /* Because the last eigenvector v(i) as theorically a null eigenvalue
   * and it might lead to inconstency values (inf() or nan()) in
   * the vector operations w(i)=A.v(i)/||A.v(i)|| and product <I.w(i)>w(i) ,
   * it is here forced to the null vector. */

  for (k = 0; k < A_nrows; k++)
    {
      *(basis + i * A_nrows + k) = 0.0;
    }


  printf ("vrification des vecteurs wi de la base\n");
  /* verification of the w(i) vectors */
  for (i = 0; i < A_ncols; i++)
    {
      for (k = i; k < A_ncols; k++)
	{
	  norme =
	    cblas_sdot (A_nrows, basis + i * A_nrows, 1, basis + k * A_nrows,
			1);
	  if (isnan (norme) == 1 || isinf (norme) == 1 || isinf (norme) == -1)
	    {
	      printf ("illegal numeric value :\n");
	      printf ("product w[%d] T. w[%d] = %3e isinf()=%d isnan()=%d\n", i,
		      k, norme, isinf (norme), isnan (norme));
	      exit (EXIT_FAILURE);
	    }
	}
    }
  return (basis);
}


/**
 * @brief Compute the image J given by :
 *        J = A.v(1) + mean
 * @param A matrix A
 * @param A_nrows number of A rows
 * @param A_ncols number of A columns
 * @param eigen_vecs eigen vector sorted in decreasing order
 * @param mean_im mean image
 */
float *
compute_image_from_v1 (float *A, int last_eigenvec, int A_nrows, int A_ncols,
		       float *eigen_vecs, float *mean_im)
{
  int i;
  float *f;

  f = (float *) calloc (A_nrows, sizeof (float));
  if (f == NULL)
    {
      printf ("allocation failed\n");
      exit (EXIT_FAILURE);
    }

  for (i = 0; i < last_eigenvec; i++)
    {
      /* product A x eigenvec(i) */
      cblas_sgemm (CblasColMajor, CblasNoTrans, CblasNoTrans,
		   A_nrows, 1, A_ncols, 1.0, A, A_nrows,
		   eigen_vecs + (i * A_ncols), A_ncols, 1.0, f, A_nrows);
    }

  /* recentering f = f + mean */
  if (mean_im != NULL)
    {
      cblas_saxpy (A_nrows, +1.0, mean_im, 1, f, 1);
    }

  return f;
}

/**
 * @brief Compute the image J given by :
 *        J = sum(i=first_vec, last_vec) <I-mu, w(i)>w(i)+ mu
 * @param A matrix A
 * @param A_nrows number of A rows
 * @param A_ncols number of A columns
 * @param eigen_vecs eigen vector sorted in decreasing order
 * @param mean_im mean image = mu
 */
float *
compute_image_from_partial_basis (float *im, int first_vec, int last_vec,
				  int npixs, float *basis_vecs, float *mean_im)
{
  int i;
  /*  int j; */
  float *J;
  float alpha;

  J = (float *) calloc (npixs, sizeof (float));
  if (J == NULL)
    {
      printf ("allocation failed\n");
      exit (EXIT_FAILURE);
    }
  if (first_vec != 0)
    first_vec--;
  for (i = first_vec; i < last_vec; i++)
    {
      /* scalar product alpha = <I-mu.w(i)>. As variable im
       comes from matrix A which is still centered by the average 
      image we only multiply im by wi */
      alpha = cblas_sdot (npixs, im, 1, basis_vecs + i * npixs, 1);

      /* J(i) = J(i-1) + alpha.w(i) */
      cblas_saxpy (npixs, +alpha, basis_vecs + i * npixs, 1, J, 1);

    }

  /* recentering J = J + mean if needed */
  if (mean_im != NULL)
    {
      cblas_saxpy (npixs, +1.0, mean_im, 1, J, 1);
    }

  return J;
}

/**
 * @brief Printf eigenvalues of matrix At.A
 *
 * @param eigenvals the eigenvalues
 * @param nvals number of eigenvalues
 */
void
print_eigenvalues (float *eigenvals, int nvals)
{
  int i;

  printf ("eigenvalues ");
  for (i = 0; i < nvals; i++)
    {
      printf ("%2.4e ", eigenvals[i]);
    }
  printf ("\n");

  printf ("eigenvalues_normalized ");
  for (i = 0; i < nvals; i++)
    {
      printf ("%2.4e ", eigenvals[i] / eigenvals[0]);
    }
  printf ("\n");

  printf ("eigenvalues_percentage ");
  float sum = 0;
  for (i = 0; i < nvals; i++)
    {
      sum += eigenvals[i];
    }
  for (i = 0; i < nvals; i++)
    {
      printf ("%2.4e ", 100 * eigenvals[i] / sum);
    }

  printf ("\n");
  
  return;
}

/**
 * @brief Printf elementary statistics about an image
 *
 * @param im the image
 * @param npixs size of the image
 */
void
print_statistics (float *im, int npixs)
{
  int i = 0;
  float min = im[0], max = im[0];
  double moyenne = 0, carre = 0, sigma = 0;

  for (i = 0; i < npixs; i++)
    {
      if (im[i] < min)
	min = im[i];
      if (im[i] > max)
	max = im[i];

      moyenne += im[i];
      carre += im[i] * im[i];
    }
  moyenne /= npixs;
  sigma = sqrt (carre / npixs - moyenne * moyenne);
  printf
    ("min %+2.4e max %+2.4e max-min %+2.4e norm %+2.4e mean %+2.4e stdev %+2.4e\n",
     min, max, max - min, sqrt (carre / npixs), moyenne, sigma);
  return;
}

float var(float * im, int npix)
{
  int i;
  float mean = 0, square = 0;
  float variance;
  for (i = 0; i < npix; i++)
    {
      mean += im[i];
      square += im[i] * im[i];
    }
  mean /= npix;
  variance = square / npix - mean * mean;
  return variance;
}

void image_image_op(float ** output, float * im1, char *operator, float * im2, int npix)
{
  int i;
  float * im3;

  if (!strcmp(operator,"$"))
    {
      im3 = *output;
      if (im3 == NULL)
	{
	  im3 = (float *) malloc(sizeof(float));
	  if (im3 == NULL)
	    {
	      perror("image_image_op()\n");
	      exit(EXIT_FAILURE);
	    }
	  *output = im3;
	}
      im3[0]=0.0;
      for (i=0; i<npix; i++) im3[0] += im1[i] * im2[i];
      return;
    }    
  
  /* allocation familiale du patch */
  im3 = *output;
  if (im3 == NULL)
    {
      im3 = (float *)malloc(npix * sizeof(float));
      if (im3 == NULL)
	{
	  perror("Problem in image_image_op()\n");
	  exit(EXIT_FAILURE);
	}
      *output = im3;
    }

  if (!strcmp(operator,"+"))
    {
      for (i=0; i<npix; i++) im3[i] = im1[i] + im2[i];
    }
  else if (!strcmp(operator,"-"))  
    {
      for (i=0; i<npix; i++) im3[i] = im1[i] - im2[i];
    }
  else if (!strcmp(operator,"*"))
    {
      for (i=0; i<npix; i++) im3[i] = im1[i] * im2[i];
    }
  else if (!strcmp(operator,"/"))
    {
      for (i=0; i<npix; i++) im3[i] = im1[i] / im2[i];
    }    
  
  return;
}

void image_scalar_op(float ** output, float * im1, char *operator, float scalar, int npix)
{
  int i;
  float * im3=*output;

  /* allocation familiale du patch */
  if (im3 == NULL)
    {
      im3 = (float *)malloc(npix * sizeof(float));
      if (im3 == NULL)
        {
          perror("Problem in image_image_op()\n");
          exit(EXIT_FAILURE);
        }
      *output = im3;
    }

  if (!strcmp(operator,"+"))
    {
      for (i=0; i<npix; i++) im3[i] = im1[i] + scalar;
    }
  else if (!strcmp(operator,"-"))  
    {
      for (i=0; i<npix; i++) im3[i] = im1[i] - scalar;
    }
  else if (!strcmp(operator,"*"))
    {
      for (i=0; i<npix; i++) im3[i] = im1[i] * scalar;
    }
  else if (!strcmp(operator,"/"))
    {
      for (i=0; i<npix; i++) im3[i] = im1[i] / scalar;
    }    

  return;
}

void zeros(float ** output, int npix)
{
  int i;
  float * im;
  
  /* allocation familiale du patch */
  im = *output;
  if (im == NULL)
    {
      im = (float *)malloc(npix * sizeof(float));
      if (im == NULL)
	{
	  perror("Problem in zeros()\n");
	  exit(EXIT_FAILURE);
	}
      *output = im;
    }

  for(i=0; i<npix; i++) im[i]=0.0f;

  return;
}
